/**
 * xrdp: A Remote Desktop Protocol server.
 *
 * MS-RDPERP : Definitions from [MS-RDPERP]
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * References to MS-RDPERP are currently correct for v20190923 of that
 * document
 */

#if !defined(MS_RDPERP_H)
#define MS_RDPERP_H

/* Window List Capability Set: WndSupportLevel (2.2.1.1.2) */
#define TS_WINDOW_LEVEL_NOT_SUPPORTED  0x00000000
#define TS_WINDOW_LEVEL_SUPPORTED      0x00000001
#define TS_WINDOW_LEVEL_SUPPORTED_EX   0x00000002


#endif /* MS_RDPERP_H */
