EXPORTS
SCardEstablishContext
SCardReleaseContext
SCardIsValidContext
SCardListReaderGroupsA
SCardListReaderGroupsW
SCardListReadersA
SCardListReadersW
SCardListCardsA
SCardListCardsW
SCardListInterfacesA
SCardListInterfacesW
SCardGetProviderIdA
SCardGetProviderIdW
SCardGetCardTypeProviderNameA
SCardGetCardTypeProviderNameW
SCardIntroduceReaderGroupA
SCardIntroduceReaderGroupW
SCardForgetReaderGroupA
SCardForgetReaderGroupW
SCardIntroduceReaderA
SCardIntroduceReaderW
SCardForgetReaderA
SCardForgetReaderW
SCardAddReaderToGroupA
SCardAddReaderToGroupW
SCardRemoveReaderFromGroupA
SCardRemoveReaderFromGroupW
SCardIntroduceCardTypeA
SCardIntroduceCardTypeW
SCardSetCardTypeProviderNameA
SCardSetCardTypeProviderNameW
SCardForgetCardTypeA
SCardForgetCardTypeW
SCardFreeMemory
SCardLocateCardsA
SCardLocateCardsW
SCardGetStatusChangeA
SCardGetStatusChangeW
SCardCancel
SCardConnectA
SCardConnectW
SCardReconnect
SCardDisconnect
SCardBeginTransaction
SCardEndTransaction
SCardState
SCardStatusA
SCardStatusW
SCardTransmit
SCardControl
SCardGetAttrib
SCardSetAttrib
